'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _lodash = require('lodash');

var _lodash2 = _interopRequireDefault(_lodash);

var _async = require('async');

var _async2 = _interopRequireDefault(_async);

var _wordpos = require('wordpos');

var _wordpos2 = _interopRequireDefault(_wordpos);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const wordpos = new _wordpos2.default();

// Unhandled promises should throw top-level errors, not just silently fail
// This is a shim for wordnet lookup.
// http://wordnet.princeton.edu/wordnet/man/wninput.5WN.html

process.on('unhandledRejection', err => {
  throw err;
});

// Does a word lookup
// @word can be a word or a word/pos to filter out unwanted types
const lookup = function lookup(word, pointerSymbol = '~', cb) {
  let pos = null;

  const match = word.match(/~(\w)$/);
  if (match) {
    pos = match[1];
    word = word.replace(match[0], '');
  }

  return wordpos.lookup(word).then(results => {
    const synets = [];

    results.forEach(result => {
      result.ptrs.forEach(part => {
        if (pos !== null && part.pos === pos && part.pointerSymbol === pointerSymbol) {
          synets.push(part);
        } else if (pos === null && part.pointerSymbol === pointerSymbol) {
          synets.push(part);
        }
      });
    });

    const itor = (word1, next) => {
      wordpos.seek(word1.synsetOffset, word1.pos).then(sub => {
        next(null, sub.lemma);
      });
    };

    _async2.default.map(synets, itor, (err, items) => {
      items = _lodash2.default.uniq(items);
      items = items.map(x => x.replace(/_/g, ' '));
      cb(err, items);
    });
  });
};

exports.default = {
  lookup
};